#!/bin/sh
################################################################################
# This script enables the lock migration feature for NFS toolkit on Linux.
# This module script is added inorder to invoke the lock migration feature
# after the IP module has been successfully started during the package startup
# and IP module is shutdown after lock migration disablement during package halt.
# This script also fixes the defect QXCR1001378534
#
# Revision History:
# 1. Removed references to the /etc/SuSE-release file. ( 28/06/2015 ) 
################################################################################
# Source utility functions.
###########################

    . /etc/cmcluster.conf
PATH=/bin:/sbin:/usr/bin:/usr/sbin:/bin:
export PATH=$PATH:$SGSBIN
NFS_FLM_SCRIPT="${TKIT_SCRIPT_DIR}/nfs.flm"
NFS_FLM_COUNTER=0
NFS_FLM_FIRST_PASS=0

# The NFS_FLM_SCRIPT variable is the name of the script that manages
# synchronizing the file lock status entries for the primary and
# adoptive nodes associated with this NFS Package.

if [[ -z $SG_UTILS ]]
then
    SG_UTILS=$SGCONF/scripts/mscripts/utils.sh
fi

if [[ -f ${SG_UTILS} ]]; then
    . ${SG_UTILS}
    if (( $? != 0 ))
    then
        echo "ERROR: Unable to source package utility functions file: ${SG_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find package utility functions file: ${SG_UTILS}"
    exit 1
fi

###################################################################
#
# Get the environment for this package through utility function
# sg_source_pkg_env().
#
###################################################################

sg_source_pkg_env $*

function verify_unsupported_online_modification
{
  typeset -i i=0
  typeset -i flag=0

  # These arrays define following things
  # curarrayname => Attribute Names for which online modifcation is not supported.
  # cuarray => Current values of the unsupported attribitues
  # prevarray => Previous values of unsupported attribtes.

  declare -a curarrayname=( "LOCK_MIGRATION" "NFS_FLM_HOLDING_DIR" "PROPAGATE_INTERVAL" )
  declare -a curarray=( "$NFS_LOCK_MIGRATION" "$NFS_NFS_FLM_HOLDING_DIR" "$NFS_PROPAGATE_INTERVAL" )
  
  #Before assigning PREV_NFS_LOCK_MIGRATION to an array convert it to upper case as we do for NFS_LOCK_MIGRATION also.
  PREV_NFS_LOCK_MIGRATION=`echo $PREV_NFS_LOCK_MIGRATION | tr "[:lower:]" "[:upper:]" | tr -d "[:space:]"` 
  declare -a prevarray=("$PREV_NFS_LOCK_MIGRATION" "$PREV_NFS_NFS_FLM_HOLDING_DIR" "$PREV_NFS_PROPAGATE_INTERVAL" )

  while (( $i <= ${#curarray[@]} -1 && $i <= ${#prevarray[@]} -1 && $i <= ${#curarrayname[@]} -1 ))
  do
   if [[ "${curarray[i]}" != "${prevarray[i]}" ]]; then
        sg_log 0 "ERROR: Online Modification of attribute ${curarrayname[i]} value is not supported"
        flag=1
   fi
   (( i = i + 1 ))
  done

 if [[ $flag -eq 1 ]]; then
   return 1
 fi

}

#########################################################################
#
# validate_command
#
# This function should define actions to take when the package is
# being validated during cmcheckconf/cmapplyconf on all the configured
# package nodes.
#
#########################################################################

function validate_command
{

   sg_log 0 "NFS Lock Migration module Validation"
   if [[ -z $SG_PRIV_CLUSTER_VERIFICATION  || $SG_COMMAND = "cmapplyconf" ]]; then
         # Check if package is applied
         $SGSBIN/cmviewcl -f line -s config -p $SG_PACKAGE > /dev/null 2>&1
         if [[ $? -eq 0 ]]; then
            # Check if package is running and check if this modules is not newly getting
            # added during online configuration, then only call function verify_unsupported_online_modification
            pkg_stat=`$SGSBIN/cmviewcl -f line | grep "$SG_PACKAGE|status" | cut -d "=" -f 2 2>&1`
            if [[ "$pkg_stat" = "up" && ${PREV_SG_MODULE_NAME[@]} = *"tkit/nfs/"* ]]; then
               verify_unsupported_online_modification
               if [[ $? -eq 1 ]]; then
                  return 1
               fi
            fi
         fi
   fi
  
    #Check if NFS_FLM_HOLDING_DIR is mentioned when
    #file lock migration is enabled.
    if [[ "$NFS_LOCK_MIGRATION" = "YES" ]]; then
       if [[ -z "$NFS_NFS_FLM_HOLDING_DIR" ]]; then
          sg_log 0 "ERROR: Holding directory ( NFS_FLM_HOLDING_DIR ) value cannot be null when lock migration is enabled"
          return 1
       fi
    fi
    return 0
}


# kill the named process
#
function killproc
{
pidlist=$(/bin/ps -o pid= -C $1)
if [[ -n "$pidlist" ]]; then
   read -a pidarray <<< $pidlist 
   for p in "${pidarray[@]}"; do
       kill -15 $p
   done
fi
sleep 1
pidlist=$(/bin/ps -o pid= -C $1)
if [[ -n "$pidlist" ]]; then
  read -a pidarray <<< $pidlist
  for p in "${pidarray[@]}"; do
     kill -9 $p
     return 1
  done
fi

}


#
# Call the script responsible for populating the sm (/var/lib/nfs/sm for SLES
# and /var/lib/nfs/statd/sm for RH  ) and holding directories with Status
# Monitor entries. For NFSv4, this script is responsible for populating the
# /var/lib/nfs/v4recovery and holding directories.
#
# Make sure the FLM script exists and is executable before launching it. Also
# kill any currently running copy of the script before launching a new copy.
#
# Pass in the command-line option of either "populate_sm_dir" to instruct the
# FLM script to copy the SM entries from the holding directory to the
# sm directory, or "populate_holding_dir" to continuously populate the
# holding directory with SM entries from sm directory.
# If we are populating the sm directory then wait for the script to complete
# and return.  If we are populating the holding directory then launch the
# script in the background as the script will not return until it is killed.
#

function call_FLM_script {

if [[ -f $NFS_FLM_SCRIPT ]]
then
        if [[ ! -x $NFS_FLM_SCRIPT ]]
        then
                chmod 544 $NFS_FLM_SCRIPT
                if [[ $? -ne 0 ]]
                then
                        sg_log 0 "$(date '+%b %e %T') - Node \"$(hostname)\": Failed to change permissions of \"$NFS_FLM_SCRIPT\".  Exiting."
                        return 1
                fi
        fi
else
        sg_log 0 "$(date '+%b %e %T') - Node \"$(hostname)\": NFS File LockMigration Script \"${NFS_FLM_SCRIPT##*/}\" does not exist, exiting."
        return 1
fi

killproc ${NFS_FLM_SCRIPT##*/}
echo "$(date '+%b %e %T') - Node \"$(hostname)\": Calling NFS FLM Script \"${NFS_FLM_SCRIPT##*/}\" with parameter \"$1\""

if [[ $1 = "populate_sm_dir" ]]
then
        ${NFS_FLM_SCRIPT} $1
        if [[ $? -eq 1 ]]; then
           sg_log 0 "ERROR: Failed to start Lock migration script"
           return 1
        fi

elif [[ $1 = "populate_holding_dir" ]]
then
        ${NFS_FLM_SCRIPT} $1 &
        if [[ $? -eq 1 ]]; then
           sg_log 0 "ERROR:Failed to start Lock migration script"
           return 1
        fi
fi
}

function determine_floating_ip
{
  # Determine the package floating IP address. If multiple floating IP's are configured,
  # then first floating IP will be used.
  PACKAGE_IP=`$SGSBIN/cmgetpkgenv $SG_PACKAGE | grep "SG_IP_ADDRESS\[0\]" | cut -d "\"" -f 2 2>/dev/null`

  if [[ ! -z $PACKAGE_IP ]]; then
    PkgIP="$PACKAGE_IP"
  else
    if [[ -x /usr/bin/host ]]; then
      NODE=`/bin/hostname`
      NODE=`/usr/bin/host -T $NODE`
      if [[ $? -eq 0 ]]; then
         NODE_IP_ADDRESS=`echo "$NODE" | cut -d " " -f 4`
         PkgIP="$NODE_IP_ADDRESS"
      else
         sg_log 0 "ERROR: /usr/bin/host command failed to determine the node IP address"
      fi
   else
      sg_log 0 "ERROR: /usr/bin/host command not found"
      sg_log 0 "ERROR: Unable to find the Node IP address"
   fi
  fi
}



#########################################################################
#
# start_command
#
# This function should define actions to take when the package starts
#
#########################################################################

function start_command
{
if [[ $NFS_LOCK_MIGRATION = "YES" ]]; then
        call_FLM_script populate_sm_dir
        if [[ $? -eq 1 ]]; then
           sg_log 0 "ERROR: Failed to start File lock migration script"
           exit 1
        else
           if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
              determine_floating_ip
              sg_log 0 "Shutting down NFS file locking daemon "
              killproc lockd
              #Applicable for both redhat and SLES 11 and above.
              sg_log 0 "Restarting rpc.statd daemon using virtual IP $PkgIP"
              killproc rpc.statd
              rpc.statd -n $PkgIP
           fi
        fi

        call_FLM_script populate_holding_dir
else
     sg_log 0 "$(date '+%b %e %T') - Node \"$(hostname)\": The parameter LOCK_MIGRATION is not set. NFS Client Locks will not be migrated."
     exit 0
fi
}

#########################################################################
#
# stop_command
#
# This function should define actions to take when the package halts
#
# It's recommended that the actions performed in this function should be
# made idempotent, that is if the halt action has already been performed
# successfully before, it should not generate an error, but exit successfully.
#
# All SG provided scripts are currently idempotent.
#
#########################################################################

function stop_command
{
if [ "${NFS_LOCK_MIGRATION}" == "YES" ]; then
    killproc  ${NFS_FLM_SCRIPT##*/}
fi
}

function validate_reconfigure_restart
{

       #Check for online modification was done to email attribute.
       sup_attri_changed=0

       # Checking NFS_FLM_RESTART and NFS_FLM_MONITOR attributes only if lock migration is enabled.
       if [[ "$NFS_LOCK_MIGRATION" = "YES" ]]; then
         if [[ "$NFS_NFS_FLM_MONITOR" != "$PREV_NFS_NFS_FLM_MONITOR" ]] ||
            [[ "$NFS_NFS_FLM_RESTART" != "$PREV_NFS_NFS_FLM_RESTART" ]]; then
               sup_attri_changed=1
         fi
       fi
       cur_ip_string=`echo ${SG_IP_ADDRESS[*]} | sed 's/ //g'`
       prev_ip_string=`echo ${PREV_SG_IP_ADDRESS[*]} | sed 's/ //g'`
       if [[ "$cur_ip_string" != "$prev_ip_string" ]]; then
          floating_ip_changed=1
          sup_attri_changed=1
       fi

        # Based on whether supported attribute changed or not and unsupported attribute
        # changed or not, take decision of restarting the service.
        # if sup_attri_changed=1 then service will be restarted
        # if sup_attri_changed=0 then service will not be restarted
        return $sup_attri_changed
}

############################################################
#
# RECONFIGURE STOP FUNCTION
#
############################################################

function reconfigure_stop_function
{
         if ! sg_check_if_reconfig_has_failed
         then
            validate_reconfigure_restart
            restart_check=$?
            if [[ $restart_check -eq 1 ]]; then
              ${SGSBIN}/cmhaltserv ${service_name}
              if [[ $? -ne 0 ]]; then
                 sg_log 999 "ERROR: Failed to stop the NFS toolkit service ( $service_name )"
                 sg_set_reconfig_failure
                 exit_val=1
              else
                 sg_log 999 "SUCCESS: Successfully Stopped NFS toolkit service $service_name"
                 sg_log 0 "SUCCESS: Successfully Stopped NFS toolkit service $service_name"
              fi
            fi
         else
            sg_log 999 "WARNING: NFS toolkit - Not attempting to stop NFS toolkit service"
         fi
}

############################################################
#
# RECONFIGURE START FUNCTION
#
############################################################
function reconfigure_start_function
{
           validate_reconfigure_restart
           restart_check=$?
           if [[ $restart_check -eq 1 ]]; then 
             if [[ "$NFS_LOCK_MIGRATION" == "YES" ]]; then 
               #If IP address changes then restart the rpc.statd if lock migration is enabled.
               if [[ $floating_ip_changed -eq 1 ]]; then
                  IP=`ps -ef | grep rpc.statd | grep -v grep | sed 's/  */\ /g' | cut -d " " -f 10`
                  if [[ "$IP" != "${SG_IP_ADDRESS[0]}" ]]; then
                     determine_floating_ip
                     if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
                          echo -e "Restarting rpc.statd daemon using virtual IP $IP"
                          killproc rpc.statd
                          rpc.statd -n $PkgIP
                     fi
                  fi
                fi
              fi

              msg="cmrunserv: Unable to start service $service_name (it is already active)"
              out=`${SGSBIN}/cmrunserv ${service_name} "${service_cmd_string}" 2>&1`
              ret=$?
              if [[ $ret -ne 0 ]] && [[ "$out" != "$msg" ]]; then
                  sg_log 999 "ERROR: Failed to start the NFS toolkit service ( $service_name )"
                  sg_set_reconfig_failure
                  exit_val=1
              elif [[ $ret -eq 0 ]]; then
                  sg_log 999 "SUCCESS: Successfully started NFS toolkit service $service_name"
                 if ! sg_check_if_reconfig_has_failed ; then
                   sg_log 0 "SUCCESS: Successfully started NFS toolkit service $service_name"
                 fi
              fi
           fi
}

function monitor_nfs_flm
{
  
while :; do
   if [[ $NFS_LOCK_MIGRATION = "YES" && $NFS_FLM_MONITOR = "YES" ]]
      then
        if [[ $NFS_FLM_FIRST_PASS -ne 0 ]]
           then
             pid=$(/bin/ps -o pid= -C ${NFS_FLM_SCRIPT##*/})
             if [[ -z "$pid" ]]
                then
                  # Couldn't locate running NFS_FLM_SCRIPT
                  # process.  Restart the FLM script unless
                  # the script has died NFS_FLM_RESTART times,
                  # in which case exit.
                  if [[ $NFS_FLM_COUNTER -ge $NFS_FLM_RESTART ]]
                     then
                       sg_log 0 "Node \"$(hostname)\" - ERROR: $NFS_FLM_SCRIPT has died too many times"
                       if [[ -n "${SG_EMAIL_ID[*]}" ]]; then
                          sg_send_mail "File Lock Migration script has gone down. NFS toolkit package $SG_PACKAGE failed on node $NODENAME."
                          sleep 3
                       fi
                       return 1
                  fi

                  sg_log 0 "WARNING: NFS toolkit - File lock migration process ( $NFS_FLM_SCRIPT ) is not running.  Restarting script $(($NFS_FLM_RESTART - $NFS_FLM_COUNTER )) more times before giving up"
                  (( NFS_FLM_COUNTER = $NFS_FLM_COUNTER + 1 ))
                  $NFS_FLM_SCRIPT populate_holding_dir &
             fi
         else
            NFS_FLM_FIRST_PASS=1
         fi
   fi
   sleep $MONITOR_INTERVAL
done
  
}

################
# MAIN routine
################

export TKIT_SCRIPT_DIR=${SGROOT}/nfstoolkit
export NFSv4_VIRTUAL_ROOT=$NFS_NFSV4_VIRTUAL_ROOT
export QUOTA_MON=$NFS_QUOTA_MON
export NFS_FLM_HOLDING_DIR=$NFS_NFS_FLM_HOLDING_DIR
export PROPAGATE_INTERVAL=$NFS_PROPAGATE_INTERVAL
export NFS_FLM_MONITOR=$NFS_NFS_FLM_MONITOR
export NFS_FLM_RESTART=$NFS_NFS_FLM_RESTART
export MONITOR_INTERVAL=$NFS_MONITOR_INTERVAL
export RETRY_INTERVAL=$NFS_RETRY_INTERVAL
export RETRY_TIMES=$NFS_RETRY_TIMES

#Convert the required parameter values to upper case.
NFS_LOCK_MIGRATION=`echo $NFS_LOCK_MIGRATION | tr "[:lower:]" "[:upper:]" | tr -d "[:space:]"`
NFS_FLM_MONITOR=`echo $NFS_FLM_MONITOR | tr "[:lower:]" "[:upper:]" | tr -d "[:space:]"`
NFS_FLM_SCRIPT="${TKIT_SCRIPT_DIR}/nfs.flm"

# This condition executes when lock migration feature is enabled. If
# user has not specified the values for NFS_FLM_RESTART and NFS_FLM_MONITOR,
# following check assigns the default values while executing the script.

if [[ $NFS_LOCK_MIGRATION = "YES" ]]; then
  if [[ -z "$NFS_FLM_RESTART" ]] ; then
     NFS_FLM_RESTART=4
  fi
  if [[ -z "$NFS_FLM_MONITOR" ]]; then
     NFS_FLM_MONITOR="NO"
  fi
fi

#Identify the service name and service command for the package
if [[ $1 = "reconfigure_stop" || $1 = "reconfigure_start" ]]; then
 if [[ ${#SG_SERVICE_CMD[@]} -ne 0 ]]; then
   i=0
   service_name=""
   service_cmd_string=""
   for cmd in "${SG_SERVICE_CMD[@]}"
   do
     echo "$cmd" | grep "lock_migration" >> /dev/null
     if [[ $? -eq 0 ]]
     then
        service_name=${SG_SERVICE_NAME[$i]}
        service_cmd_string="$cmd"
        break
     fi
     i=$i+1
   done
 fi
fi

typeset -i exit_val=0

case ${1} in
     start)
           start_command $*
           exit_val=$?
           ;;

     stop)
           stop_command $*
           exit_val=$?
           ;;

     validate)
           validate_command $*
           exit_val=$?
           ;;
     monitor)
           monitor_nfs_flm
           exit_val=$?
           ;;
    
     reconfigure_stop)
           reconfigure_stop_function
           ;;

     reconfigure_start)
           reconfigure_start_function
           ;; 

     *)
           sg_log 0 "Unknown entry point $1"
           ;;
esac

exit $exit_val

