#!/bin/bash
###################################################################
# (C) Copyright 2007-2015 Hewlett-Packard Development Company, L.P.
# Source utility functions.
# Version: A.12.00.30
# Revision History
#  Date
#   28/08/2012               1. Modification was made to support online
#                               modification of package attributes. 
#   04/10/2013               1. Modification was made to support Cluster
#                               verification feature for NFS toolkit
#                            2. Online modification defect fixes are
#                               introduced. 
#   11/29/2013               1. Removed NFS toolkit maintenance mode support.
#                            2. Changes to obsolete legacy style of packaging. 
#                            3. Added changes to check if previous module has failed during
#                               online modification of toolkit attributes and online
#                               addition of toolkit modules. 
#   24/07/2014               1. Added changes to export XFS elements individually (QXCR1001366201).
#   29/10/2014               1. Removed validation for lock migration attribute and
#                               moved to lock_migration.sh script ( QXCR1001378534 )
#   29/06/2015               1. Removed references to the /etc/SuSE-release file to support
#				SLES 12 OS.
################################################################################
. /etc/cmcluster.conf
PATH=/bin:/sbin:/usr/bin:/usr/sbin:/bin:
export PATH=$PATH:$SGSBIN

if [[ -z $SG_UTILS ]]; then
        SG_UTILS=$SGCONF/scripts/mscripts/utils.sh
fi

if [[ -f ${SG_UTILS} ]]; then
    . ${SG_UTILS}
    if (( $? != 0 ))
    then
        echo "ERROR: Unable to source package utility functions file: ${SG_UTILS}"
        exit 1
    fi
else
    echo "ERROR: Unable to find package utility functions file: ${SG_UTILS}"
    exit 1
fi

###################################################################
#
# Get the environment for this package through utility function
# sg_source_pkg_env().
#
###################################################################

sg_source_pkg_env $*

##################################################################
#
# VALIDATE FUNCTION
#
#################################################################

function validate_function
{
      sg_log 0 "NFS Toolkit validation"

      # SG version check to decide to enable/disable CV feature and online modification
      # validation checks.
      if [[ "$sg_version" > "11.20.20" || "$sg_version" = "11.20.20" ]];  then
       HOST=`/bin/hostname| cut -d "." -f 1`
       # Check if the SG_COMMAND executed is "cmcheckconf -P" and package is up and running"
       if [[ -z $SG_PRIV_CLUSTER_VERIFICATION  || $SG_COMMAND = "cmapplyconf" ]]; then
         # Check if package is applied
            $SGSBIN/cmviewcl -f line -s config -p $SG_PACKAGE > /dev/null 2>&1
            if [[ $? -eq 0 ]]; then
            # Check if package is running and check if this modules is not newly getting
            # added during online configuration, then only call function verify_unsupported_online_modification
            pkg_stat=`$SGSBIN/cmviewcl -f line | grep "$SG_PACKAGE|status" | cut -d "=" -f 2 2>&1`
             if [[ "$pkg_stat" = "up" && ${PREV_SG_MODULE_NAME[@]} = *"tkit/nfs"* ]]; then
                verify_unsupported_online_modification
                if [[ $? -eq 1 ]]; then
                   return 1 
                fi
             fi
            fi
        fi

      
        #Toolkit verificaion of NFS toolkit begins here

        # Check for the presence of lockfile command which is required during
        # rmtab synchronization process.
        if [[ ! -f /usr/bin/lockfile ]]; then
          sg_log 0 "ERROR: lockfile command does not exist" 
          exit_code=1
        fi

        # Check for the presence of portmap in redhat 6
        if [[ -f /etc/redhat-release ]]; then
           release=`cat /etc/redhat-release | cut -d " " -f 7 | cut -d "." -f 1`
           if [[  $release -eq 5  ]]; then
             portmap_path=`which portmap 2> /dev/null`
             if [[ ! -f $portmap_path ]]; then
               sg_log 0 "ERROR: portmap binary does not exist"
               exit_code=1
             fi
           fi
        else #if sles system, then check for the path of rpcbind
           rpcbind_path=`which rpcbind 2>> /dev/null`
           if [[ ! -x $rpcbind_path ]]; then
              sg_log 0 "ERROR: rpcbind binary does not exist"
              exit_code=1
           fi
        fi 

        # Since quotamon daemon is optional, we will check the presence only if this rquotad
        # daemon is explicitly enabled in package configuration file.
        if [[ "$NFS_QUOTA_MON" = "yes" || "$NFS_QUOTA_MON" = "YES" ]]; then
         rquota_path=`which rpc.rquotad 2> /dev/null`
         if [[ ! -f $rquota_path ]]; then
           sg_log 0 "ERROR: rpc.rquotad binary does not exist"
           exit_code=1
         fi
        fi
 
        #Check the presence of application binaries that are common in SLES and REDHAT 
        typeset -i m=0 
        declare -a common_app_bin=( "rpc.nfsd" "rpc.mountd" "rpc.statd" )
        while (( $m <= ${#common_app_bin[@]} -1 ))
        do
               temp=`which ${common_app_bin[m]} 2> /dev/null`
               if [[ ! -f $temp ]]; then
                 sg_log 0 "ERROR: RPC service binary ( ${common_app_bin[m]} ) does not exist"
                 exit_code=1
               fi
               (( m = $m + 1 ))
        done 

         # Check for the presence of NFSv4 specific daemons
         if [[ -n "$NFS_NFSV4_VIRTUAL_ROOT" ]]; then
            idmap_path=`which rpc.idmapd 2> /dev/null`
            if [[ ! -x $idmap_path ]]; then
               sg_log 0 "ERROR: rpc.idmapd binary does not exist"
               exit_code=1
            fi
         fi 
  

       # Following part of verification happens during cmcheckconf -v
       # SG_PRIV_CLUSTER_VERIFICATION variable is provided by SG to identify whether cmcheckconf -v is used.
       # if SG_PRIV_CLUSTER_VERIFICATION is "y" then cmcheckconf -v is being executed.
       # if SG_PRIV_CLUSTER_VERIFICATION is null then cmcheckconf -p is being executed.
       if [[ "$SG_PRIV_CLUSTER_VERIFICATION" = "y" ]]; then
         if [[ -n "$NFS_NFSV4_VIRTUAL_ROOT" ]]; then

         typeset -i vrflag=0
         
         #Check if there is any virtual root already defined in nodes for any NFSv4 communicaton. 
         showmount -e > /dev/null 2>&1
         if [[ $? -eq 0 ]]; then
          showmount -e | grep "fsid=0" >> /dev/null
           if [[ $? -eq 0 ]]; then
             vrflag=1
           fi
         fi

         if [[ -f /etc/exports ]]; then
           grep "fsid=0" /etc/exports >> /dev/null
           if [[ $? -eq 0 ]]; then
             vrflag=1
           fi  
         fi

         if [[ $vrflag -eq 1 ]]; then
           sg_log 0 "WARNING: NFS toolkit - Virtual root is already defined in the $HOST"
         fi
        fi
       fi

      if [[ $exit_code -eq 1 ]]; then
        exit 1
      fi
     
     fi # End of SG version check for 11.20.10
     
     return $exit_code
}


#######################################################################
#
# VERIFY UNSUPPORTED ONLINE MODIFICATION FUNCTION
# This function will get invoked only on SG version A.11.20.20 or later
#######################################################################

function verify_unsupported_online_modification
{
  typeset -i i=0
  typeset -i flag=0

  # These arrays define following things
  # curarrayname => Attribute Names for which online modifcation is not supported. 
  # cuarray => Current values of the unsupported attribitues
  # prevarray => Previous values of unsupported attribtes.

  declare -a curarrayname=( "NFSV4_VIRTUAL_ROOT" )
  declare -a curarray=( "$NFS_NFSV4_VIRTUAL_ROOT" )
  declare -a prevarray=( "$PREV_NFS_NFSV4_VIRTUAL_ROOT" )

  while (( $i <= ${#curarray[@]} -1 && $i <= ${#prevarray[@]} -1 && $i <= ${#curarrayname[@]} -1 ))
  do
   if [[ "${curarray[i]}" != "${prevarray[i]}" ]]; then
        sg_log 0 "ERROR: Online Modification of attribute ${curarrayname[i]} value is not supported"
        flag=1
   fi
   (( i = i + 1 ))
  done

 # Verify if any XFS attribute value is added,modified or deleted.
 cur_xfs=`echo ${NFS_XFS[*]} | sed 's/ //g'`
 prev_xfs=`echo ${PREV_NFS_XFS[*]} | sed 's/ //g'`

  if [[ "$cur_xfs" != " " || "$prev_xfs" != " " ]]; then
    if [[ "$cur_xfs" != "$prev_xfs" ]]; then
         flag=1
         sg_log 0 "ERROR: Online modification of XFS attribute value is not supported" 
    fi
  fi

 if [[ $flag -eq 1 ]]; then
   return 1
 fi

}

#############################################################
#
# START FUNCTION
#
############################################################

function start_function
{

   if sg_check_if_reconfig_has_failed ; then
      sg_log 999 "WARNING: NFS toolkit - Not attempting to start the newly added NFS toolkit module"
      return 
   fi

   ${TKIT_SCRIPT_DIR}/toolkit.sh start
   exit_code=$?
   if sg_check_if_it_is_reconfiguration ; then
     if [[  $exit_code -ne 0 ]]; then
       sg_log 999 "ERROR: Failed to start the newly added NFS toolkit module"
       sg_set_reconfig_failure
     else
       sg_log 999 "SUCCESS: Successfully started newly added NFS toolkit module"
       sg_log 0 "SUCCESS: Successfully started newly added NFS toolkit module"
     fi
   fi

}

#############################################################
#
# STOP FUNCTION
#
############################################################

function stop_function
{
        ${TKIT_SCRIPT_DIR}/toolkit.sh stop
        exit_code=$?
}


#############################################################
#
# MONITOR FUNCTION
#
############################################################

function monitor_function
{
        ${TKIT_SCRIPT_DIR}/nfs.mon
        exit_code=$?
}

############################################################
#
# RECONFIGURE STOP FUNCTION 
#
############################################################


function reconfigure_stop_function
{
         if ! sg_check_if_reconfig_has_failed 
         then  
            validate_reconfigure_restart
            restart_check=$?
            if [[ $restart_check -eq 1 ]]; then
              ${SGSBIN}/cmhaltserv ${service_name}
              if [[ $? -ne 0 ]]; then
                 sg_log 999 "ERROR: Failed to stop the NFS toolkit service ( $service_name )" 
                 sg_set_reconfig_failure 
                 exit_code=1
              else
                 sg_log 999 "SUCCESS: Successfully Stopped NFS toolkit service $service_name"
                 sg_log 0 "SUCCESS: Successfully Stopped NFS toolkit service $service_name"
              fi
            fi
         else
            sg_log 999 "WARNING: NFS toolkit - Not attempting to stop NFS toolkit service"
         fi
}

############################################################
#
# RECONFIGURE START FUNCTION 
#
############################################################
function reconfigure_start_function
{
           validate_reconfigure_restart
           restart_check=$?
           if [[ $restart_check -eq 1 ]]; then
              msg="cmrunserv: Unable to start service $service_name (it is already active)"
              out=`${SGSBIN}/cmrunserv ${service_name} "${service_cmd_string}" 2>&1`
              ret=$?
              if [[ $ret -ne 0 ]] && [[ "$out" != "$msg" ]]; then
                  sg_log 999 "ERROR: Failed to start the NFS toolkit service ( $service_name )"
                  sg_set_reconfig_failure 
                  exit_code=1
              elif [[ $ret -eq 0 ]]; then
                  sg_log 999 "SUCCESS: Successfully started NFS toolkit service $service_name"
                 if ! sg_check_if_reconfig_has_failed ; then
                   sg_log 0 "SUCCESS: Successfully started NFS toolkit service $service_name"
                 fi
              fi
           fi
}


################################################################
#
# VALIDATE RECONFIGURE RESTART FUNCTION
#
################################################################

function validate_reconfigure_restart
{

        #Check for online modification was done to email attribute.
        sup_attri_changed=0
       
          var1=`echo ${SG_EMAIL_ID[*]} | sed 's/ //g'`
          var2=`echo ${PREV_SG_EMAIL_ID[*]} | sed 's/ //g'`
            if [[ "$var1" != "$var2" ]]; then
              sup_attri_changed=1
          fi
 
        #Checking if other supported attributes are changed or not
        if [[ "$NFS_QUOTA_MON" != "$PREV_NFS_QUOTA_MON" ]] || 
        [[ "$NFS_MONITOR_INTERVAL" != "$PREV_NFS_MONITOR_INTERVAL" ]] || 
        [[ "$NFS_RETRY_INTERVAL" != "$PREV_NFS_RETRY_INTERVAL" ]] || 
        [[ "$NFS_RETRY_TIMES" != "$PREV_NFS_RETRY_TIMES" ]] || 
        [[ "$SG_SCRIPT_LOG_FILE" != "$PREV_SG_SCRIPT_LOG_FILE" ]]; then
          sup_attri_changed=1
        fi

        # Based on whether supported attribute changed or not and unsupported attribute
        # changed or not, take decision of restarting the service.
        # if sup_attri_changed=1 then service will be restarted
        # if sup_attri_changed=0 then service will not be restarted

        return $sup_attri_changed
}




##################################
#
# START OF ROUTINES
#
#################################

typeset -i exit_code=0

############################################
# Get the SG version
###########################################

sg_version=`$SGSBIN/cmversion | cut -d "." -f 2-4`

export TKIT_SCRIPT_DIR=$SGROOT/nfstoolkit
export NFSv4_VIRTUAL_ROOT=$NFS_NFSV4_VIRTUAL_ROOT
export QUOTA_MON=$NFS_QUOTA_MON
export LOCK_MIGRATION=$NFS_LOCK_MIGRATION
export NFS_FLM_HOLDING_DIR=$NFS_NFS_FLM_HOLDING_DIR
export PROPAGATE_INTERVAL=$NFS_PROPAGATE_INTERVAL
export MONITOR_INTERVAL=$NFS_MONITOR_INTERVAL
export RETRY_INTERVAL=$NFS_RETRY_INTERVAL
export RETRY_TIMES=$NFS_RETRY_TIMES

# Exporting array elements individually
# as it is required to be processed in hanfs.sh  
typeset -i k=0
for k in ${!NFS_XFS[*]}
do
 # export command does not accepts XFS[] format for exporting
 # variable. Hence exporting as XFS_* to hanfs.sh script.
 export XFS_$k="${NFS_XFS[$k]}"
done

#Identify the service name and service command for the package
if [[ $1 = "reconfigure_stop" || $1 = "reconfigure_start" ]]; then
 if [[ ${#SG_SERVICE_CMD[@]} -ne 0 ]]; then
   i=0
   service_name=""
   service_cmd_string=""
   for cmd in "${SG_SERVICE_CMD[@]}"
   do
     echo "$cmd" | grep "nfs_monitor" >> /dev/null
     if [[ $? -eq 0 ]]
     then
        service_name=${SG_SERVICE_NAME[$i]}
        service_cmd_string="$cmd"
        break
     fi
     i=$i+1
   done
 fi
fi

case $1 in
        validate)
                validate_function
                ;;

        start)
                start_function
                ;;

        stop)
                stop_function
                ;;
                
        nfs_monitor)
        
                monitor_function
                ;;

        reconfigure_stop)

                reconfigure_stop_function
                ;;

        reconfigure_start)

                reconfigure_start_function
                ;;

        *)
                sg_log 0 "Usage: ${0} [ start | stop | validate | nfs_monitor | reconfigure_stop | reconfigure_start ]"
                ;;
esac

exit $exit_code
