#!/bin/bash
# (C) Copyright 2007-2016 Hewlett Packard Enterprise Development LP.
# Version: A.12.00.51
#
# This script contains Serviceguard NFS toolkit defined environment
# variables and functions that are used for following:
# 1. Starting and Stopping NFS daemons.
# 2. Exporting and Unexporting NFS shares.
# 3. Starting and Stopping rmtab synchronization process.
# 4. Starting and Stopping File lock migration process.
#
# Revision History:
# 1) Modification was made to support the file lock migration feature.
# 2) Modification was made to support the version 4 of the NFS protocol 
#    for NFS server.( 09/02/2012 ) 
# 3) Modification was made to merge the lock migration script functionalities
#    as a part of startup based on lock migration attributes. ( 18/03/2012 )
# 4) Modification was made to merge the sync_rmtab binary functionalities 
#    into hanfs.sh script in the form of shell script functions. sync_rmtab
#    binary file has been removed. ( 28/08/2012 )
# 5) Added fix for the quix QXCR1001265600 and QXCR1001265598 ( 10/04/2013 )
# 6) Added fix for the defects QXCR1001366201 and QXCR10013662 ( 24/07/2014 ) 
# 7) Added fix for the defect QXCR1001375119 ( 14/10/2014 )
# 8) Added fix for the defect QXCR1001378534 ( 28/10/2014 ) 
# 9) Removed references to the /etc/SuSE-release file with support for sles12 OS. ( 29/06/2015 )
# 10) Added fix for the defect QXCR1001435955 ( 21/01/2016 )
# 11) Added fix for the defect QXCR1001436124 ( 27/01/2016 )
# Exit value definition:
# 1 - a failure occurs in one of the function.
#     
#Include the SG PATHs

. /etc/cmcluster.conf

PATH=/bin:/sbin:/usr/bin:/usr/sbin:/bin:

# Importing XFS variables from the tkit_module.sh
# Constructing new array called XFS using the values of XFS_ array.
for i in ${!XFS_*}
do
  XFS[${i#XFS_}]="${!i}"
done

#Define array for holding sorted XFS values
declare -a SORTED_XFS

# Define number of NFS listener daemons to be launched
RPCNFSDCOUNT=8


#Sourcing variables from nfstkit_utils.sh

NFSUTILS="$SGROOT/nfstoolkit/nfstkit_utils.sh"

if [[ -f $NFSUTILS ]]; then
    . $NFSUTILS
    if [[ $? -ne 0 ]]; then
        echo "ERROR:Could not source nfs utility script"
        exit 1
    fi
else
    echo "ERROR:Could not find nfs utility script"
    exit 1
fi


# START OF RUN FUNCTIONS

# Definitions for script internal use. Do not modify this part.
NFS_RMTAB_FILE=/var/lib/nfs/rmtab
SYNC_RMTAB_PID_FILE="$TKIT_SCRIPT_DIR/sync_rmtab.PID.$SG_PACKAGE"
NFS_EXTAB_LOCK=/var/lock/extab.lock
QUOTA_MON=`echo $QUOTA_MON | tr "[:lower:]" "[:upper:]" | tr -d "[:space:]"`

# Locking mechanism for restarting nfsd,rpc.statd and rpc.lockd during
# package halt in a no lock migration configuration.
FLM_LOCKDIR="/tmp/nfslock"
FLM_LOCKFILE="/tmp/nfslock/nfslock.$$"
FLM_LOCK="/tmp/nfslock*"


# START OF RUN FUNCTIONS

function get_temp_lock
{

# LOCK_WAIT specifies the number of seconds to wait for
# acquiring the lock.
#
LOCK_WAIT=120
LOCK_COUNTER=0

while (( $LOCK_COUNTER < $LOCK_WAIT ))
do
        mkdir $FLM_LOCKDIR 2>/dev/null
        if [[ $? -eq 0 ]]
        then
                touch $FLM_LOCKFILE
                break
        fi
        (( LOCK_COUNTER = $LOCK_COUNTER + 1 ))
        sleep 1
done

if [[ $LOCK_COUNTER -eq $LOCK_WAIT ]]
then
        echo "$(date '+%b %e %T') - Node \"$(hostname)\": could not get temporary lock within $LOCK_WAIT seconds"
        exit 1
fi
}

function release_temp_lock
{
rm -rf $FLM_LOCK 2>/dev/null
if [[ $? -ne 0 ]]; then
   echo -e "$(date '+%b %e %T') - Node \"$(hostname)\": Failed to remove the lock"
else
   echo -e "$(date '+%b %e %T') - Node \"$(hostname)\": Successfully removed the temporary lock"
fi
}

signal_temp_lock_cleanup()
{
if [[ -f $FLM_LOCKFILE ]]
then
        release_temp_lock
        echo "$(date '+%b %e %T') - Node \"$(hostname)\": Received signal and temporary lock is removed"
fi
        exit 1
}

# Handle signal to remove the lock dir
trap signal_temp_lock_cleanup HUP INT QUIT TERM



# This function starts the NFS daemons
#
function nfs_defined_run_cmds
{
    echo "$(date '+%b %e %T') - Node \"$(hostname)\": Starting NFS daemons."
    nfsd_flag=0
    # start each NFS daemon if it was not running
    if [[ -z ${NFSv4_VIRTUAL_ROOT} ]] ; then
       if [[ $release -eq 0 ]] || [[ ! -z "$IS_SUSE" ]]; then 
          rpcbind_pid=`pidofproc rpcbind`
          if [[ -z $rpcbind_pid ]]; then
             verify_nfs_daemons_before_rpcbind portmap rpc.statd rpc.rquotad rpc.mountd nfsd lockd
             echo -e "\tStarting rpcbind daemon"
             if [[ -f $rpc_service ]]; then
                systemctl start rpcbind  
             else
                /sbin/rpcbind
             fi
             check_return 16
          else
             echo -e "\trpcbind had been running on this node."
             verify_nfs_daemons_after_rpcbind portmap rpc.statd rpc.rquotad rpc.mountd nfsd lockd
          fi
       else 
          pidlist=`pidofproc portmap`
          if [[ -z $pidlist ]]; then
             echo -e "\tStarting portmap daemon"
             portmap
             check_return 16
          else
             echo -e "\tportmap had been running on this node."
          fi
       fi

        #This is a Red Hat or Suse 11 and above system, then run rpc.statd 
        pidlist=`pidofproc rpc.statd`
        if [[ -z $pidlist ]]; then     # execute if not already running
            echo -e "\tStarting rpc.statd daemon"
            if [[ -f $statd_service ]]; then
                systemctl start rpc-statd
            else
                rpc.statd
            fi
            check_return 16
        else
            echo -e "\trpc.statd had been running on this node."
        fi

    if [ $QUOTA_MON == "YES" ]; then
        pidlist=`pidofproc rpc.rquotad`
        if [[ -z $pidlist ]]; then
            echo -e "\tStarting rpc.rquotad daemon"
            rpc.rquotad
            check_return 16
        else
            echo -e "\trpc.rquotad had been running on this node."
        fi
    fi
  fi

    # Only following daemons are needed for NFSv4. If package is 
    # configured for NFSv4, then following daemons will be started.
    # If package is configured for NFSv3/NFSv2, along with above daemons,
    # below daemons will also be started. 

    #rpcbind is needed for SLES 11 and above as it supports ipv6.
   if [[ ! -z "${IS_SUSE}" ]] && [[ ! -z ${NFSv4_VIRTUAL_ROOT} ]]; then
      pidlist=`pidofproc rpcbind`
      if [[ -z $pidlist ]]; then
          verify_nfs_daemons_before_rpcbind nfsd rpc.mountd rpc.idmapd 
          echo -e "\tStarting rpcbind daemon"
          if [[ -f $rpc_service ]]; then
              systemctl start rpcbind  
          else
              /sbin/rpcbind
          fi
          check_return 16
      else
          echo -e "\trpcbind had been running on this node."
          verify_nfs_daemons_after_rpcbind nfsd rpc.mountd rpc.idmapd
      fi
 
   fi 

    if [[ -f $nfs_service ]] && [[ ! -z $IS_SUSE ]]; then
        env_file="/etc/sysconfig/nfs"
        cp -f $env_file $env_file.bkp 2>/dev/null
        sed -i 's/USE_KERNEL_NFSD_NUMBER.*/USE_KERNEL_NFSD_NUMBER=\"8\"/' $env_file
    fi
	
    nfsd_flag=0
    pidlist=`pidofproc nfsd`
    #if NFSD has not been running then, in SLES12 and RH7 environments
    #we check systemcl status. If status shows it is running then we 
    #use systemctl restart. 
    if [[ -z $pidlist ]]; then
        echo -e "\tStarting $RPCNFSDCOUNT nfsd daemons"
        if [[ -f $nfs_service ]]; then
            nfsd_flag=1
            systemctl status nfs-server | grep -iq "Active: active"
            retval=$?
            if [[ $retval -eq 0 ]]; then
                systemctl restart nfs-server
                if [[ $? -ne 0 ]]; then
                    echo "$(date '+%b %e %T') - Node \"$(hostname)\":ERROR: Failed to start nfsd daemons"
                    exit 1
                fi  
            else
                systemctl start nfs-server 
                if [[ $? -ne 0 ]]; then
                    echo "$(date '+%b %e %T') - Node \"$(hostname)\":ERROR: Failed to start nfsd daemons"
                    exit 1
                fi
            fi
        else
            rpc.nfsd $RPCNFSDCOUNT
        fi
        check_return 16
    else
        echo -e "\tnfsd had been running on this node."
    fi
	
    pidlist=`pidofproc rpc.mountd`
    if [[ -z $pidlist ]]; then
        echo -e "\tStarting rpc.mountd daemon"
        if [[ -f $mountd_service ]]; then
            nfsd_flag=1
            systemctl start nfs-mountd
        else
            rpc.mountd
        fi
        check_return 16
        sleep 1
    else
        echo -e "\trpc.mountd had been running on this node."
    fi

    nfsd_flag=0
    if [[ -n ${NFSv4_VIRTUAL_ROOT} ]]; then
       pidlist=`pidofproc rpc.idmapd`
       if [[ -z $pidlist ]]; then
         echo -e "\tStarting rpc.idmapd daemon"
         if [[ -f idmap_service ]]; then
             nfsd_flag=1
             systemctl start nfs-idmapd
         else
             rpc.idmapd  
         fi
         check_return 16
       else
         echo -e "\trpc.idmapd had been running on this node"
       fi
    fi
    
    if [[ -f $nfs_service ]] && [[ ! -z $IS_SUSE ]]; then
        mv -f $env_file.bkp $env_file 2>/dev/null
    fi     
}


# This function kills NFS daemons.
#
function nfs_defined_halt_cmds
{
# it is not necessary to stop NFS daemons. They may be started on boot time
# or currently used by other packages. Inorder to ensure successful
# unmount of the exported file systems, only nfsd, rpc.statd and lockd daemons
# will be restarted during package halt when lock migration is not enabled.
nfsd_flag=0
LOCK_MIGRATION=`echo $LOCK_MIGRATION | tr "[:lower:]" "[:upper:]" | tr -d "[:space:]"`
if [ $LOCK_MIGRATION == "YES" ]; then
    if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
      #Release all the locks to unmount the filesystem successfully
      if [[ $release -eq 1 ]]; then
       killproc lockd
      fi

      if [[ ${QUOTA_MON} == "YES" ]]; then
        echo -e "\tStopping rpc.rquotad daemon"
       killproc rpc.rquotad
      fi

      # for Redhat machine or Suse 11 or above system, restart statd
         echo -e "\tRestarting statd daemon"
         killproc rpc.statd
         if [[ -f $statd_service ]]; then
             systemctl start rpc-statd
         else
             rpc.statd
         fi
    else
       echo -e "\tStopping idmapd daemon"
       killproc rpc.idmapd
    fi

    echo -e "\tStopping rpc.mountd daemon"
    killproc rpc.mountd
    echo -e "\tStopping nfsd daemon"
    

    #in case of NFSD in RH7 and SLES12 we'll have to do systemctl stop
    #instead of killproc as next start will fail if we do killproc

    killproc nfsd
    if [[ -f $nfs_service ]]; then
        systemctl stop nfs-server
    fi

    # After the upgrade to the 12.00.20 version of NFS toolkit, the package created with
    # older version should succesfully halt during cmhaltpkg or during failover.
    # Below code helps to halt the nfs.flm script of the package created with the previous
    # version of NFS toolkit package. This ensures successful halt of the old package
    # after upgrading.
    echo ${SG_OPERATION_SEQUENCE[@]} | grep "lock_migration.sh"
    if [[ $? -ne 0 ]]; then
       killproc nfs.flm
    fi
else
    
    if [[ ! -f $nfs_service ]]; then
        # Function get_temp_lock helps to get the temporary lock before restarting the
        # statd, nfsd and lockd daemons. This is to ensure that, at a time
        # only one NFS package restart these daemons.
        get_temp_lock
        echo -e "$(date '+%b %e %T') - Node \"$(hostname)\": Restarting nfsd,statd and lockd daemons to allow successful unmount of file system"
        #Stopping the rpc.statd, lockd and nfsd daemons to avoid the umount busy issue

        if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
            killproc rpc.statd
            statd_pidlist=`pidofproc rpc.statd`
            if [[ ! -z $statd_pidlist ]]; then 
                echo -e "$(date '+%b %e %T') - Node \"$(hostname)\": WARNING: Failed to stop the rpc.statd"
            fi
            if [[ $release -eq 1 ]]; then
                killproc lockd
                lockd_pidlist=`pidofproc lockd`
                if [[ ! -z $lockd_pidlist ]]; then
                    echo -e "$(date '+%b %e %T') - Node \"$(hostname)\": WARNING: Failed to stop the rpc.lockd"
                fi 
            fi
        fi
        killproc nfsd 
        nfsd_pidlist=`pidofproc nfsd`
        if [[ ! -z $nfsd_pidlist ]]; then
            echo -e "$(date '+%b %e %T') - Node \"$(hostname)\": WARNING: Failed to stop the nfsd"
        fi

        # Restarting the statd,lockd and nfsd daemons after stopping.
        if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
            rpc.statd
            check_return 22
        
            if [[ $release -eq 1 ]]; then
               rpc.lockd
               check_return 22
            fi
        fi
        rpc.nfsd $RPCNFSDCOUNT
        check_return 22
        
        # Releasing the temporary lock aquired in the beginning of this block. 
        release_temp_lock
    fi
fi 

}


# This function checks if the XFS is a logic volume mounted file system.
#
function check_lv_fs
{

    #Function modified for using awk instead of grep for more specific search
    # of mount VGs.
    # Grep will give confict with vgNFS1 & vgNFS11
    mountinfo="$1"
    result=`mount | awk '$3 == filesys {print $1}' filesys="$mountinfo"`
    echo $result
}

function sort_exported_directories
{
    typeset -i i=0
    typeset -i a=0
    declare -a indexes
    declare -a num_of_comp
    typeset Y
    typeset X
    typeset Z
    typeset dirnm
    typeset temp
    typeset temp1
    typeset -i j

    while (( $a < ${#XFS[*]} ))
    do
        (( indexes[a] = $a ))

        #seperate export options from export directory
         Z="${XFS[a]}"
         Z="${Z%/}"
         Z="${Z##*:}"
         Y=$Z
        #Make sure all the mnt points end with a / for uniformity
        if [[ ${Y%/} = $Y ]]
        then
            Y=$Y/
        fi
        dirnm=$Y

        # Calculate number of components in the moint point path
        (( num_of_comp[a] = $(echo $dirnm | awk -F "/" '{print NF}') ))
        (( a = a + 1 ))
    done

    ((i=0))
    ((temp=0))
    ((temp1=0))

    #sort num_of_comp[*] and save the indexes
    while ((i<${#num_of_comp[*]}))
    do
       ((j=i+1))
        while ((j < ${#num_of_comp[*]}))
        do
            if (( ${num_of_comp[i]} > ${num_of_comp[j]} ))
            then
                temp=${num_of_comp[i]}
                num_of_comp[i]=${num_of_comp[j]}
                num_of_comp[j]=$temp

                temp1=${indexes[i]}
                indexes[i]=${indexes[j]}
                indexes[j]=$temp1
            fi
            ((j=j+1))
        done
        ((i=i+1))
    done

    for ((i=0;i<${#indexes[@]};i++))
    do
      SORTED_XFS[${#SORTED_XFS[*]}]="${XFS[${indexes[i]}]}"
    done
}




# For each exported file system, export it with the options
# using exportfs -i(1m).
#
function export_fs
{
    I=""
    for I in "${SORTED_XFS[@]}"
    do
        while :
        do
            #Processing the XFS attribute value in the format required for exporting
            I="${I%/}"
            exp_dir_name="${I##*:}"
            exp_opt="${I%:*}"
            
            echo "$(date '+%b %e %T') - Node \"$(hostname)\": Exporting filesystem on ${exp_dir_name}"
            ISMOUNTED=`check_lv_fs ${exp_dir_name}`
            if [[ -z $ISMOUNTED ]]; then
                echo -e "\tWARNING: NFS toolkit - export directory ( ${exp_dir_name} ) is not mounted on a volume group"
            fi

            # Lock the exportfs table to prevent multiple packages to export/unexport
            # file systems at the same time.
            lockfile -l 15 $NFS_EXTAB_LOCK
            EXPORTERROR=`exportfs -i ${exp_opt}:${exp_dir_name} 2>&1`
            # unlock the exportfs table
            rm -f $NFS_EXTAB_LOCK

            XTAB_IN_USE=`echo $EXPORTERROR | grep "xtab in use"`
            if [[ -n "$XTAB_IN_USE" ]]; then
                echo -e "\tWARNING: NFS toolkit - exportfs -i ${exp_opt}:${exp_dir_name}: $EXPORTERROR. Retrying in 2 seconds"
                sleep 2
            else
                if [[ -n "$EXPORTERROR" ]]; then
                    echo -e "\t$EXPORTERROR"
                    # unlock the exportfs table
                    rm -f $NFS_EXTAB_LOCK
                    [[ 1 == 0 ]]
                    check_return 17
                fi
                break
            fi
        done
    done
}

# This function exports only NFSv4 supported export directories.
# Export and unexport mechanism is different in NFSv4, hence  
# exporting/unexporting is not clubbed with NFSv3/NFSv2 type
# of exporting/unexporting.

function export_fs_nfsv4
{
I=""
#Create NFSV4 directory if it does not exist.
if [[ ! -d ${NFSv4_VIRTUAL_ROOT} ]]; then
  mkdir -p $NFSv4_VIRTUAL_ROOT
  chmod 755 $NFSv4_VIRTUAL_ROOT
fi


#Exporting the virtual root directory. For virtual root directory, fsid should be 0.
#Having Multiple virtual roots is not supported as it causes unexpected behavior by NFS.
exportfs -i -o fsid=0 *:${NFSv4_VIRTUAL_ROOT}
if [[ $? -ne 0 ]]; then
  echo -e "\tERROR: Failed to export NFSv4 virtual root directory ( ${NFSv4_VIRTUAL_ROOT} )"
  check_return 17
fi

for I in "${SORTED_XFS[@]}"
    do
        while :
        do
            #Processing the XFS attribute value to get export directory and export options.
            I="${I%/}"
            exp_dir_name="${I##*:}"
            exp_opt="${I%:*}"
            mkdir -p ${NFSv4_VIRTUAL_ROOT}${exp_dir_name}
            mount --bind ${exp_dir_name} ${NFSv4_VIRTUAL_ROOT}${exp_dir_name} 
 
            echo "$(date '+%b %e %T') - Node \"$(hostname)\": Exporting filesystem on ${exp_dir_name}"
            ISMOUNTED=`check_lv_fs ${exp_dir_name}`
            if [[ -z $ISMOUNTED ]]; then
                echo -e "\tWARNING: NFS toolkit - export directory ( ${exp_dir_name} ) is not mounted on a volume group"
            fi

            # Lock the exportfs table to prevent multiple packages to export/unexport
            # file systems at the same time.

            lockfile -l 15 $NFS_EXTAB_LOCK
            EXPORTERROR=`exportfs -i ${exp_opt}:${NFSv4_VIRTUAL_ROOT}${exp_dir_name} 2>&1`
            # unlock the exportfs table
            rm -f $NFS_EXTAB_LOCK
            XTAB_IN_USE=`echo $EXPORTERROR | grep "xtab in use"`
            if [[ -n "$XTAB_IN_USE" ]]; then
                echo -e "\tWARNING: NFS toolkit - exportfs -i ${NFSv4_VIRTUAL_ROOT}${exp_dir_name}: $EXPORTERROR. Retrying in 2 seconds"
                sleep 2
            else
                if [[ -n "$EXPORTERROR" ]]; then
                    echo -e "\t$EXPORTERROR"
                    # unlock the exportfs table
                     rm -f $NFS_EXTAB_LOCK
                    [[ 1 == 0 ]]
                    check_return 17
                fi
                break
            fi
        done
    done
}
 
                    

# Unexport file system
#
function un_export_fs
{
    I=""
    for ((i=${#SORTED_XFS[@]}-1;i>=0;i--)) 
    do
        I="${SORTED_XFS[$i]}"
        while :
        do

            #Processing the XFS attribute value in the format required for unexporting
            I="${I%/}"
            exp_dir_name="${I##*:}"
 
            echo "$(date '+%b %e %T') - Node \"$(hostname)\": Unexporting filesystem on ${exp_dir_name}"
            # Lock the exportfs table to prevent multiple packages to export/unexport
            # file systems at the same time.
            lockfile -l 15 $NFS_EXTAB_LOCK
            EXPORTERROR=`exportfs -u $I 2>&1`
            # unlock the exportfs table
            rm -f $NFS_EXTAB_LOCK

            XTAB_IN_USE=`echo $EXPORTERROR | grep "xtab in use"`
            if [[ -n "$XTAB_IN_USE" ]]; then
                echo -e "\tWARNING: NFS toolkit - exportfs -u $I: $EXPORTERROR. Retrying in 2 seconds"
                sleep 2
            else
                if [[ -n "$EXPORTERROR" ]]; then
                    echo "$EXPORTERROR"
                    [[ 1 == 0 ]]
                    check_return 18
                fi
                break
            fi
        done
    done
}

# This function unexports only NFSv4 exported directories.

function un_export_fs_nfsv4
{
   I=""
   
    # Nested shares would cause problems while unbinding
    # after unexporting. Hence unexporting nfsv4 shares
    # in the reverse order of FS shares which are binded  
    # and exported in the beginnng. 
    # Ex: export order : /mnttest, /mnttest/test1
    #     unexport order : /mnttest/test1, /mnttest
    # This is to ensure successful umount/unbind is done for
    # nfsv4 FS shares during the halt of the package.
    # Note: This is not required for nfsv2/v3 fs shares, as FS does not
    # required to be "bind" before exporting.

    for ((i=${#SORTED_XFS[@]}-1;i>=0;i--))
    do
        I="${SORTED_XFS[$i]}"
        while :
        do

            #Processing the XFS attribute value in the format required for unexporting
            I="${I%/}"
            exp_dir_name="${I##*:}"
            exp_opt="${I%:*}"
 
            echo "$(date '+%b %e %T') - Node \"$(hostname)\": Unexporting filesystem on $exp_dir_name"

            # Lock the exportfs table to prevent multiple packages to export/unexport
            # file systems at the same time.

            lockfile -l 15 $NFS_EXTAB_LOCK
            EXPORTERROR=`exportfs -u ${exp_opt}:${NFSv4_VIRTUAL_ROOT}${exp_dir_name} 2>&1`
            # unlock the exportfs table
            rm -f $NFS_EXTAB_LOCK
            
			#Unmounting the virtual root directories 
			#to ensure that successful unbinding is done during shutdown.
			umount ${NFSv4_VIRTUAL_ROOT}${exp_dir_name} 2>/dev/null
			if [[ $? -eq 0 ]]; then
				echo -e "\tNFS toolkit - Unmounting exported directory ( ${NFSv4_VIRTUAL_ROOT}${exp_dir_name} ) is successful"
				rm -rf ${NFSv4_VIRTUAL_ROOT}${exp_dir_name}
			else
				echo -e "\tERROR: NFS toolkit - Failed to unmount exported directory ( ${NFSv4_VIRTUAL_ROOT}${exp_dir_name} )" 
				#If unmount fails we should not remove the directory as the actual directory contents will also be deleted.
				echo -e "\tWARNING: Ensure that the directory ( ${NFSv4_VIRTUAL_ROOT}${exp_dir_name} ) is unmounted and its contents are deleted before restarting the package"
			fi

            XTAB_IN_USE=`echo $EXPORTERROR | grep "xtab in use"`
            if [[ -n "$XTAB_IN_USE" ]]; then
                echo -e "\tWARNING: NFS toolkit - exportfs -u ${exp_opt}:${NFSv4_VIRTUAL_ROOT}${exp_dir_name}: $EXPORTERROR. Retrying in 2 seconds"
                sleep 2
            else
              if [[ -n "$EXPORTERROR" ]]; then
                    echo "$EXPORTERROR"
                    [[ 1 == 0 ]]
                    check_return 18
                fi
                break
            fi
        done
    done

   # Check if this virtual root is used by other packages in the system. If yes
   # do not unexport it.
   /bin/mount -v | /bin/grep "${NFSv4_VIRTUAL_ROOT}" >> /dev/null
   if [[ $? -ne 0 ]] ; then
     exportfs -u *:${NFSv4_VIRTUAL_ROOT}
   fi
}

# Find all pid of a process
#
function pidofproc
{
   pid_list=$(/bin/ps -o pid= -C $1)
   echo $pid_list 
}


# kill the named process
#
function killproc
{
    pidlist=`pidofproc $1`
    if [[ -n "$pidlist" ]]; then 
        read -a pidarray <<< $pidlist
        for pid in "${pidarray[@]}"; do
            kill -15 $pid
        done
    fi    
    sleep 1
    pidlist=`pidofproc $1`     
    if [[ -n "$pidlist" ]]; then 
       read -a pidarray <<< $pidlist
       for pid in "${pidlist[@]}"; do  
          kill -9 $pid
          check_return 21
       done 
    fi    
}

###########################################
#Function verify_nfs_daemons_before_rpcbind
#
#This function is used to check whether
#RPC/NFS daemons are running already
#before starting rpcbind daemon.
############################################

function verify_nfs_daemons_before_rpcbind
{
    for var in "$@"
    do
        pidlist=`pidofproc $var`
        if [[ ! -z $pidlist ]]; then
           echo -e "Stopping the daemon $var as it is not registered with the currently running rpcbind"
           if [[ "$var" = "nfsd" ]] && [[ -f $nfs_service ]]; then
               systemctl stop nfs-server
           else
               killproc $var
           fi
        fi
    done
}

###########################################
#Function verify_nfs_daemons_after_rpcbind
#
#This function is used to check whether
#RPC/NFS daemons are started before rpcbind
#has started by pinging it using rpcinfo.
############################################

function verify_nfs_daemons_after_rpcbind
{
     for var in "$@"
     do
         I=0
         while [ $I -lt ${#HA_SERVER[@]} ]; do
             if [ "$var" == "${HA_SERVER_PROC[$I]}" ]; then
                 rpcinfo -u $HOST ${HA_SERVER_PROG[$I]} ${HA_SERVER_VER[$I]}
                 if [[ $? -ne 0 ]]; then
                     if [[ "$var" = "nfsd" ]] && [[ -f $nfs_service ]]; then
                         systemctl stop nfs-server
                     else
                         killproc $var
                     fi
                 fi
             fi
             let I="$I + 1"
         done
     done
}



# Check return value of functions and exit.
#
function check_return
{
    if (( $? != 0 ))
    then
        case $1 in
            16)
                echo "ERROR: Function nfs_defined_run_cmds"
                echo "ERROR: Failed to start a NFS daemon"
                nfs_defined_halt_cmds
                if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
                  un_export_fs
                  synchronize_rmtab export
                else
                  un_export_fs_nfsv4
                fi
                if [[ $nfsd_flag -eq 1 ]]; then
                    mv -f $env_file.bkp $env_file 2>/dev/null
                fi 
                exit 1
                ;;

            17)
                if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then
                  echo "ERROR: Function export_fs"
                  echo "ERROR: Failed to export $I"
                  un_export_fs
                  synchronize_rmtab export
                else
                   echo "ERROR: Function export_fs_nfsv4"
                   echo "ERROR: Failed to export $I" 
                   un_export_fs_nfsv4
                fi 
                exit 1
                ;;

            18)
                if [[ -z ${NFSv4_VIRTUAL_ROOT} ]]; then 
                  echo "ERROR: Function un_export_fs"
                else
                  echo "ERROR: Function un_export_fs_nfsv4"
                fi
                echo "ERROR: Failed to unexport $I"
                exit_value=1 
                ;;

            21)
                echo "ERROR: Function nfs_defined_halt_cmds"
                echo "ERROR: Failed to stop nfs daemon"
                exit_value=1
                ;;

            22)
                echo "ERROR: Function nfs_defined_halt_cmds"
                echo "ERROR: Failed to restart rpc daemons during package halt"
                if [[ $nfsd_flag -eq 1 ]]; then
                    mv -f $env_file.bkp $env_file 2>/dev/null
                fi
                exit_value=1
                ;; 

            31)
                echo "ERROR: Function stop_sync_rmtab_bg_proc"
                echo "ERROR: Failed to stop sync rmtab background process"
                exit_value=1
                ;;
            *)
                echo "ERROR: Failed, unknown error."
                ;;
        esac
    fi
}

# This function will be used by the rmtab synchronization
# process for locking rmtab file, sync file and temp file
# while performing the operations like import, export and
# sync of the files.

function lock_file
{
if [[ -x /usr/bin/lockfile ]];then
  /usr/bin/lockfile -l 15 $1.lock
  if [[ $? -ne 0 ]]; then
     echo "ERROR: lock_file: Failed to lock the file $1"
     exit 1
  fi
else
  echo "ERROR: lock_file: lockfile command does not exist"
  exit 1
fi
}

# This Function will be used by the rmtab synchronization
# process for unlocking rmtab file, sync file and temp file
# after operations like write and read are complete.

function unlock_file
{

if [[ -f $1.lock ]]; then
  rm -f $1.lock
fi

}

# This function will copy the contents from the sync file in the
# shared storage to the /var/lib/nfs/rmtab file. This function uses
# directory mentioned in the XFS[0] attribute value to store the sync file.

function import_rmtab
{

if [[ ! -f $SYNC_FILE ]]; then
  echo "import_rmtab: $SYNC_FILE does not exist"
  return 0
fi

lock_file $RMTAB_PATH
rsync $SYNC_FILE $RMTAB_PATH 
if [[ $? -ne 0 ]]; then
  echo "WARNING: NFS toolkit - import_rmtab failed to copy the contents from sync file in shared storage to /var/lib/nfs/rmtab file"
  unlock_file $RMTAB_PATH
  return 1
fi
unlock_file $RMTAB_PATH
}

# This function will sync the contents from the /var/lib/nfs/rmtab to
# the sync file in the shared storage.

function sync_rmtab_file
{

if [[ ! -f $SYNC_FILE ]]; then
  echo "sync_rmtab_file: $SYNC_FILE does not exist"
fi

lock_file $RMTAB_PATH
lock_file $SYNC_FILE
if [[ -x /usr/bin/rsync ]]; then
  rsync $RMTAB_PATH $SYNC_FILE
  if [[ $? -ne 0 ]]; then
    echo "ERROR: sync_rmtab_file: Failed to sync the contents from rmtab file to sync_file"
    unlock_file $RMTAB_PATH
    unlock_file $SYNC_FILE
    return 1
  fi
fi
unlock_file $RMTAB_PATH
unlock_file $SYNC_FILE

}


# This function will alter the contents of /var/lib/nfs/rmtab file.
# Since we have to modify the /var/lib/nfs/rmtab file contents,
# its safer to use temporary file instead of directly operating
# on /var/lib/nfs/rmtab. Hence temporary file will be used.

function alter_rmtab
{

lock_file $RMTAB_PATH
lock_file $TEMP_FILE

rsync $RMTAB_PATH $TEMP_FILE
if [[ $? -ne 0 ]]; then
  echo "ERROR: alter_rmtab: Copying contents from /var/lib/nfs/rmtab file to temp file failed"
  return 1
fi
rsync $TEMP_FILE $RMTAB_PATH
if [[ $? -ne 0 ]]; then
  echo "ERROR: alter_rmtab: Copying the contents from the temp file to /var/lib/nfs/rmtab file failed"
  return 1
fi
unlock_file $RMTAB_PATH
unlock_file $TEMP_FILE

}

# This function will copy the client mount entries from /var/lib/nfs/rmtab to
# sync file which is in shared storage. This function also removes the entries from the
# /var/lib/nfs/rmtab.

function export_rmtab
{

sync_rmtab_file
if [[ $? -ne 0 ]]
then
  echo "ERROR: export_rmtab: syncing rmtab file failed"
  return 1
fi

alter_rmtab
if [[ $? -ne 0 ]]
then
 echo "ERROR: export_rmtab: altering rmtab file failed"
 return 1
fi

}

# This function will be called by synchronize_rmtab function. 

function sync_rmtab_function {

case $1 in
    import)
          import_rmtab
          return $?
          ;;

    sync)
           sync_rmtab_file
           return $? 
           ;;

    export)

           export_rmtab
           return $?
           ;;
    *)
           echo "sync_rmtab_function: Usage: sync_rmtab import|sync|export exported-file-system"
           return 1
           ;;
esac

}

# This function is provided for the rmtab file synchronization purpose.
# It can be called for exporting clients current mounted information to a cluster
# shared storage. It can be called for importing most recent client mounted information
# from the shared storage to the local rmtab file. It also can be called for refreshing
# the sync file which stored in the shared storage.
# The main goal of this function is to achieve clients seamless failover on
# NFS file system mount and unmount.
#
# This function takes one parameter (export, import, or sync)
# Note: NFSv4 does not populates rmtab file with client entries. So
# This function is not required to be called for NFSv4 server packages.
#
function synchronize_rmtab
{
    local temp=${XFS[0]}
    local dest=${temp##*:}


    # Check if an exported shared storage exist
    if [[ -z $dest ]]
    then
        echo "ERROR: sync rmtab: No exported file system on this package."
        return 1
    fi

   
    #Define the paths of the files used by rmtab synchronization process
    dest=`df -P $dest | tail -1 | awk '{print $6}'` 
    RMTAB_PATH="/var/lib/nfs/rmtab"
    SYNC_FILE_PATH=$dest
    SYNC_FILE=${SYNC_FILE_PATH}/.rmtab_sync
    TEMP_FILE="/var/lib/nfs/rmtab_temp"
 

    # Check if the source/destination directory is a mounted volume group
    if [[ -z "$dest" ]]
    then
        echo "ERROR: sync rmtab: exported file system is not a mounted volume group."
        return 1
    fi


    sync_rmtab_function $1 
    return $?

}


# Periodically check if there is a new client mount point. Update
# the rmtab syncronization file if there is a change.
# This function will be called and run as a background task
# Note: NFSv4 does not populates rmtab file with client entries. So
# This function is not required to be called for NFSv4 server packages.

function sync_rmtab_periodically
{

        local timestamp_prev="";
        local timestamp_curr="";

        while :; do
                timestamp_curr=`ls -l --time-style=full-iso $NFS_RMTAB_FILE | awk '{print $6$7}'`;
                if [[ $timestamp_curr != $timestamp_prev ]]
                then
                        synchronize_rmtab sync
                        if (( $? != 0 ))
                        then
                                exit 1;
                        fi
                fi
                timestamp_prev=$timestamp_curr;
                sleep 2;
        done
}


# Start synchronizing rmtab background process
#
# Note: NFSv4 does not populates rmtab file with client entries. So
# This function is not required to be called for NFSv4 server packages.

function start_sync_rmtab_bg_proc
{
    # This function assume that each HA package maintains its own directory.
    # i.e. no package shares directory. Otherwise this method will
    # be messed up.
    echo "$(date '+%b %e %T') - Node \"$(hostname)\": Starting rmtab synchronization process"
    sync_rmtab_periodically &
    echo $! > $SYNC_RMTAB_PID_FILE
}


# Stop synchronizing rmtab background process
#
# Note: NFSv4 does not populates rmtab file with client entries. So
# This function is not required to be called for NFSv4 server packages.

function stop_sync_rmtab_bg_proc
{
    if [[ -f $SYNC_RMTAB_PID_FILE ]]; then
        syncrmtabpid=`cat $SYNC_RMTAB_PID_FILE`
        if [[ -n $syncrmtabpid ]]; then
            echo "$(date '+%b %e %T') - Node \"$(hostname)\": Stopping rmtab synchronization process"
            kill -9 $syncrmtabpid
            check_return 31
            rm -f $SYNC_RMTAB_PID_FILE
        else
            echo "WARNING: NFS toolkit - Stopping rmtab synchronization proces: $SYNC_RMTAB_PID_FILE file corrupted"
        fi
    else
        echo "WARNING: NFS toolkit - Stopping rmtab synchronization proces: $SYNC_RMTAB_PID_FILE does not exist"
    fi
}

#-------------------MAINLINE NFS Code Starts Here-----------------
#
# FUNCTION STARTUP SECTION.  HV

exit_value=0
typeset syncrmtabpid

Action=`echo $1 | awk -F":" '{print $1}'`
set $Action

#Sort XFS directories
sort_exported_directories

# Test to see if we are being called to run the package, or halt the package.
if [[ $Action = "start" ]]
then
    
    if [[ -z ${NFSv4_VIRTUAL_ROOT} ]];then
       synchronize_rmtab import
       if (( $? != 0 ))
       then
         exit 1
       fi
       nfs_defined_run_cmds
       export_fs
       start_sync_rmtab_bg_proc
    else
       nfs_defined_run_cmds
       export_fs_nfsv4
    fi
    
elif [[ $Action = "stop" ]]
then

#rpcbind is needed for a clean shutdown in sles11 and above
    if [[ ! -z ${IS_SUSE} ]]; then
      echo "Starting rpcbind to ensure the clear shutdown of other daemons"
      if [[ -f $rpc_service ]]; then
         systemctl start rpcbind
      else
         service rpcbind start
      fi
    fi

    if [[ -z ${NFSv4_VIRTUAL_ROOT} ]];then
      stop_sync_rmtab_bg_proc
      un_export_fs
      nfs_defined_halt_cmds
      synchronize_rmtab export
      if (( $? != 0 ))
      then
        exit_value=1
      fi
    else
       un_export_fs_nfsv4
       nfs_defined_halt_cmds
    fi

fi

exit $exit_value
